# -*- coding: utf-8 -*-
""" Created on Sun Sep  5 07:13:23 2021 @authors: JWB/TRM"""


from __future__ import print_function
from ortools.linear_solver import pywraplp
import time
import xlwings as xw
import pandas as pd
import itertools
import os

#%%  Monitoring of time
class ticToc():
    r""" Class for assesing the time consumption """
    
    def __init__(self,ticTocName=''):
        self.tic = time.time()
        self.ticTocName = ticTocName
    
    def toc(self):                
        print(self.ticTocName+' calculation time: '+'%.3f seconds.'%(time.time()-self.tic))

#%% Creating model structure 

class expando(object):
    r""" A small class which can have attributes set """
    pass

class BalanceOptimizer():
    r""" 
    This class creates a model for estimating the cost 
    for supplying the hydrogen, heat and ancillirary service demands.
    """
    
    def __init__(self):
        r"""
        Creates the initial structure for data, variables, constraints and sets.
        """
        self.data   = expando()
        self.var    = expando()
        self.cts    = expando()
        self.sets   = expando()
        self.U      = {}
        
        
    def optimize(self):
        r""" Runs the optimization """
        self.modelStatus = self.model.Solve()
        self.status() # Returns model status
    
    def status(self):
        r""" Checks and prints the model status. """
        # Check model status
        if self.modelStatus == pywraplp.Solver.OPTIMAL or self.modelStatus == pywraplp.Solver.FEASIBLE:
            print('\nSolution:')
            print('Objective value, MDKK =', self.model.Objective().Value())  # Called outside, BalOpt.model.Objective().Value()      
            print('Number of variables =', self.model.NumVariables())
            print('Number of constraints =', self.model.NumConstraints())
            
            for u in uSet:
                print('Investment Capacities '+u+' = [MW]',         round(BalOpt.U[u].var.InvCapacity.solution_value()))
            print('Investment Capacities H2 storage [MWh] = ',      round(BalOpt.var.InvestH2Storage.solution_value()))
            print('Investment Capacities H2 storage [MW] = ',       round(BalOpt.var.InvestH2Discharge.solution_value()))
            print('Investment Capacities Battery storage [MWh] = ', round(BalOpt.var.InvestBatStorage.solution_value()))
            print('Investment Capacities Battery storage [MW] = ',  round(BalOpt.var.InvestBatDischarge.solution_value()))
            print('Investment Capacities Grid Connection [MW] = ',  round(BalOpt.var.InvestGridConnect.solution_value()))
        
        else:
            print('\nThe problem does not have an optimal solution.')  
               
    def load_data(self,zoneData,zoneSetup,unitData,ancS):
        r""" Reads and interprets input data """

        self.data.zoneData      = zoneData.to_dict()
        self.data.zoneSetup     = zoneSetup
        self.data.unitData      = unitData.to_dict()
        self.data.ancService    = ancS.to_dict()
        
    def load_sets(self,tSet,uSet,aSet):
        r""" Reads global sets """
        self.sets.t = tSet   # Timestep
        self.sets.u = unitData.loc['Capacity'][abs(unitData.loc['Capacity'])>=0].index.tolist()# Unit
        self.sets.a = aSet   # Each ancillary service
        
        
    def build_model(self):
        r"""
        Builds an LP model by using underlying functions.
        """        
        self.model = pywraplp.Solver('simple_mip_program', pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)      #Or use .CLP_LINEAR_PROGRAMMING
        

        # Loops through each unit to create unit variables and constraints
        for uID in self.sets.u:
            print('Creating '+uID)
            self._buildZoneUnits(uID,self.data.unitData,self.data.zoneData,self.data.zoneSetup,self.data.ancService)
            
        #Energy balance constraints        
        self._buildEnergyBalance(uID,self.data.zoneData,self.data.zoneSetup)
        
        # Objective function        
        self._build_objective()
        
    def _buildZoneUnits(self,uID,unitData,zoneData,zoneSetup,ancService):
        r"""
        This function creates a new unit with the name UiD with the given parameters from the 
        Excel Sheet (AvgUnit and Avaiability from timeline).
        The function is called from the "Build_Model"
        """
        self.U[uID]      = expando()
        self.U[uID].var  = expando()
        self.U[uID].cts  = expando()                                    
        self.U[uID].sets = expando()
        self.U[uID].data = expando()

        m = self.model        
        
        self.U[uID].data.ZoneBounds = {}

        ## -------- Build variable bounds --------
        self.U[uID].data.ZoneBounds[('pUnit',uID,'lb')] = min(self.data.unitData[uID]['Capacity'],0)
        self.U[uID].data.ZoneBounds[('pUnit',uID,'ub')] = max(self.data.unitData[uID]['Capacity'],0) 
        self.U[uID].data.ZoneBounds[('qUnit',uID,'lb')] = 0
        self.U[uID].data.ZoneBounds[('qUnit',uID,'ub')] = self.data.unitData[uID]['bPQ'] + (self.data.unitData[uID]['aPQ']) * self.data.unitData[uID]['Capacity']
        
        ## -------- Build variables --------   
        self.U[uID].var.totalCosts  = {}
        self.U[uID].var.pCosts      = {}
        self.U[uID].var.qCosts      = {}   
        self.U[uID].var.pNet        = {}
        self.U[uID].var.qNet        = {}
        self.U[uID].var.ASRevenue   = {}        
        self.U[uID].var.ancService  = {}
        self.U[uID].var.InvCapacity = m.NumVar(lb=abs(self.data.unitData[uID]['CapacityMin']),ub=abs(self.data.unitData[uID]['Capacity']),name=f"{uID}.var.InvCapacity")
           
        ##  -------- Hourly variables -------- 
        for t in self.sets.t:                              
            
            self.U[uID].var.totalCosts[t]  = m.NumVar(lb=-m.infinity(), ub= m.infinity(), name=f"{uID}.var.totalCosts.t{t}") 
            self.U[uID].var.ASRevenue[t]   = m.NumVar(lb=-m.infinity(), ub= m.infinity(), name=f"{uID}.var.ASRevenue.t{t}")                                         
            self.U[uID].var.pCosts[t]      = m.NumVar(lb=-m.infinity(), ub= m.infinity(), name=f"{uID}.var.pCosts.t{t}")
            self.U[uID].var.qCosts[t]      = m.NumVar(lb=-m.infinity(), ub= m.infinity(), name=f"{uID}.var.qCosts.t{t}")                                  
            
            # if self.U[uID] == "Battery":
            #     self.U[uID].var.pNet[t]    = m.NumVar(lb= -self.data.unitData[uID]['Capacity'] * self.data.zoneData[uID][t], ub= self.data.unitData[uID]['Capacity'] * self.data.zoneData[uID][t], name=f"{uID}.var.pNet.t{t}")                
            # else: 
            self.U[uID].var.pNet[t]        = m.NumVar(lb= min(self.data.unitData[uID]['Capacity'] * self.data.zoneData[uID][t],0), ub=max(self.data.unitData[uID]['Capacity'] * self.data.unitData[uID]['Power'],0) * self.data.zoneData[uID][t], name=f"{uID}.var.pNet.t{t}")                    
            self.U[uID].var.qNet[t]        = m.NumVar(lb= 0, ub= self.U[uID].data.ZoneBounds[('qUnit',uID,'ub')] * self.data.zoneData[uID][t], name=f"{uID}.var.qNet.t{t}")


        ##  -------- Hourly variables for each ancillirary service --------             
            self.U[uID].var.ancService[t]  = {}            
            for AncS in self.sets.a:
                self.U[uID].var.ancService[t][AncS] = m.NumVar(lb=0, ub=abs(self.data.unitData[uID]['Capacity']) * self.data.ancService[uID][AncS] * self.data.zoneData[uID][t], name=f"{uID}.var.ancService.t{t}.{AncS}")       

            
        ## -------- Build constraints --------          
        self.U[uID].cts.totalCosts            = {}                                     
        self.U[uID].cts.pCosts                = {}
        self.U[uID].cts.qCosts                = {}
        self.U[uID].cts.pNet                  = {}
        self.U[uID].cts.pGross                = {}
        self.U[uID].cts.qNet                  = {}
        self.U[uID].cts.qNet2                 = {}
        self.U[uID].cts.downReg               = {}
        self.U[uID].cts.upReg                 = {}
        self.U[uID].cts.pHourRampUp           = {}
        self.U[uID].cts.pHourRampDw           = {}   
        self.U[uID].cts.ASRevenue             = {}
        self.U[uID].cts.ASProvided            = {}
        self.U[uID].cts.InvestmentConstraint  = {}
      
        
        ## -------- Hourly constraints --------                 
        for t in self.sets.t:
            self.U[uID].cts.InvestmentConstraint[t]   = m.Add( self.U[uID].var.pNet[t] * self.data.unitData[uID]['regDirection'] <= self.U[uID].var.InvCapacity * self.data.zoneData[uID][t], name=f"{uID}.cts.InvConstraint.t{t}")               
            self.U[uID].cts.totalCosts[t]             = m.Add( self.U[uID].var.totalCosts[t] == self.U[uID].var.pCosts[t]+ self.U[uID].var.qCosts[t], name=f"{uID}.cts.totCosts.t{t}")           
            self.U[uID].cts.pCosts[t]                 = m.Add( self.U[uID].var.pCosts[t] == -self.U[uID].var.pNet[t] * (self.data.zoneData['Elspot'][t] + self.data.zoneData['Heat_Price'][t] * self.data.unitData[uID]['aPHeat']), name=f"{uID}.cts.pCosts.t{t}")            
            self.U[uID].cts.qCosts[t]                 = m.Add( self.U[uID].var.qCosts[t] == self.U[uID].var.qNet[t] *  self.data.unitData[uID]['qPOT'], name=f"{uID}.cts.qCosts.t{t}")   
            self.U[uID].cts.ASRevenue[t]              = m.Add( self.U[uID].var.ASRevenue[t] == sum(self.U[uID].var.ancService[t][AncS]*self.data.zoneData[AncS+'_Price'][t] for AncS in self.sets.a), name=f"{uID}.cts.ASRevenue.t{t}") 
            self.U[uID].cts.qNet[t]                   = m.Add( self.U[uID].var.qNet[t] == self.U[uID].var.pNet[t] * self.data.unitData[uID]['aPQ'], name=f"{uID}.cts.qNet.t{t}")      
            
            for AncS in self.sets.a:
                self.U[uID].cts.ASProvided[t]         = m.Add( self.U[uID].var.ancService[t][AncS] <= self.data.ancService[uID][AncS] * self.data.zoneData[uID][t] * self.U[uID].var.InvCapacity, name=f"{uID}.cts.ASProvided.AncS{AncS}.t{t}") 
                         
            if self.data.unitData[uID]['regDirection'] > 0:
                self.U[uID].cts.downReg[t]            = m.Add(self.U[uID].var.pNet[t] >= sum(self.U[uID].var.ancService[t][AncS] * self.data.ancService['Dwn'][AncS] for AncS in self.sets.a), name=f"{uID}.cts.downReg.t{t}")    
                self.U[uID].cts.upReg[t]              = m.Add(self.U[uID].var.pNet[t] <= self.U[uID].var.InvCapacity * self.data.zoneData[uID][t] - sum(self.U[uID].var.ancService[t][AncS] * self.data.ancService['Up'][AncS] for AncS in self.sets.a), name=f"{uID}.cts.upReg.t{t}")
            else:
                self.U[uID].cts.downReg[t]            = m.Add(self.U[uID].var.pNet[t] >= -self.U[uID].var.InvCapacity + sum(self.U[uID].var.ancService[t][AncS] * self.data.ancService['Dwn'][AncS] for AncS in self.sets.a), name=f"{uID}.cts.downReg.t{t}")   
                self.U[uID].cts.upReg[t]              = m.Add(self.U[uID].var.pNet[t] <= -sum(self.U[uID].var.ancService[t][AncS] * self.data.ancService['Up'][AncS] for AncS in self.sets.a), name=f"{uID}.cts.upReg.t{t}")
            
    # -------- Hourly unit constraints from hour 2 and forth --------   
            if t > 0:
                self.U[uID].cts.pHourRampUp[t]       = m.Add(self.U[uID].var.pNet[t] - self.U[uID].var.pNet[t-1] <= abs(self.data.unitData[uID]['Capacity']) * self.data.zoneData[uID][t] * self.data.unitData[uID]['RampeUp_h'], name=f"{uID}.cts.pHourRampUp.t{t}")
                self.U[uID].cts.pHourRampDw[t]       = m.Add(-self.U[uID].var.pNet[t] + self.U[uID].var.pNet[t-1] <= abs(self.data.unitData[uID]['Capacity']) * self.data.zoneData[uID][t] * self.data.unitData[uID]['RampDown_h'], name=f"{uID}.cts.pHourRampDw.t{t}")                
            

    def _buildEnergyBalance(self,uID,zoneData,zoneSetup):
        r""" 
        Builds the energy balances for hydrogen, power consumption in Point of connection, thus including max limits
        on import/export while having the possibility of "producing behind the meter". 
        """  

        ## -------- Import Data --------   
        m = self.model
               
        ## -------- Build variables  --------    
        self.var.Storage                        = {}
        self.var.StorageCharge                  = {}
        self.var.StorageDischarge               = {}
        self.var.BatStorage                     = {}
        self.var.BatStorageCharge               = {}
        self.var.BatStorageDischarge            = {}
        self.var.H2Sales                        = {}
        self.var.Import                         = {}
        self.var.ElBalance                      = {}
        self.var.BatteryStorage                 = {}
        self.var.Curtailment                    = {}
        self.var.InvestH2Charge                 = m.NumVar(lb=-int(zoneSetup.loc['StorageSize_MW']),      ub=-int(zoneSetup.loc['StorageSizeMin_MW']),     name=f"var.InvestH2Charge")
        self.var.InvestH2Discharge              = m.NumVar(lb=int(zoneSetup.loc['StorageSizeMin_MW']),    ub=int(zoneSetup.loc['StorageSize_MW']),         name=f"var.InvestH2Discharge")
        self.var.InvestH2Storage                = m.NumVar(lb=int(zoneSetup.loc['StorageSizeMin_MWh']),   ub=int(zoneSetup.loc['StorageSize_MWh']),        name=f"var.InvestH2Storage")
        self.var.InvestBatCharge                = m.NumVar(lb=-int(zoneSetup.loc['BatStorageSize_MW']),   ub=-int(zoneSetup.loc['BatStorageSizeMin_MW']),  name=f"var.InvestBatCharge")
        self.var.InvestBatDischarge             = m.NumVar(lb=int(zoneSetup.loc['BatStorageSizeMin_MW']), ub=int(zoneSetup.loc['BatStorageSize_MW']),      name=f"var.InvestBatDischarge")
        self.var.InvestBatStorage               = m.NumVar(lb=int(zoneSetup.loc['BatStorageSizeMin_MWh']),ub=int(zoneSetup.loc['BatStorageSize_MWh']),     name=f"var.InvestBatStorage")
        self.var.InvestGridConnect              = m.NumVar(lb=int(zoneSetup.loc['GridConnectMin_MW']),    ub=int(zoneSetup.loc['GridConnect_MW']),         name=f"var.InvestGridConnect")
        self.var.ancServiceBat                  = {}
        self.var.ASRevenueBat                   = {}

        
        ## -------- Build Constraints  --------    
        self.cts.H2Balance                      = {}
        self.cts.StorageBal                     = {}
        self.cts.Importmin                      = {}
        self.cts.BatSoCBal                      = {}
        self.cts.ElBalance                      = {}
        self.cts.InvConstraintH2                = {}
        self.cts.InvConstraintH2Charge1         = {} 
        self.cts.InvConstraintH2Charge2         = {} 
        self.cts.InvConstraintH2Charge3         = {}
        self.cts.InvConstraintH2Charge4         = {}
        self.cts.InvConstraintBat               = {} 
        self.cts.InvConstraintBatCharge1        = {} 
        self.cts.InvConstraintBatCharge2        = {} 
        self.cts.InvConstraintBatCharge3        = {} 
        self.cts.InvConstraintBatCharge4        = {} 
        self.cts.InvConstraintGC1               = {}
        self.cts.InvConstraintGC2               = {}
        self.cts.AncillaryServices1             = {}
        self.cts.AncillaryServices2             = {}       
        self.cts.downRegBat                     = {}
        self.cts.upRegBat                       = {}
        self.cts.ASProvidedBat                  = {}
        self.cts.ASRevenueBat                   = {}
        
            # -------- Investment Constraints   --------        
        self.cts.InvConstraintH2Charge4         = m.Add(self.var.InvestH2Discharge  <= self.var.InvestH2Storage,  name=f"cts.InvConstraintH2Charge4")
        self.cts.InvConstraintBatCharge4        = m.Add(self.var.InvestBatDischarge <= self.var.InvestBatStorage, name=f"cts.InvConstraintBatCharge4")


            # -------- Hourly variable bounds       --------                  
        for t in self.sets.t:
            self.var.Storage[t]                 = m.NumVar(lb=0, ub=int(zoneSetup.loc['StorageSize_MWh']),      name=f"var.Storage.t{t}")
            self.var.StorageCharge[t]           = m.NumVar(lb=-int(zoneSetup.loc['StorageSize_MW']), ub=0,      name=f"var.StorageCharge.t{t}")
            self.var.StorageDischarge[t]        = m.NumVar(lb=0, ub=int(zoneSetup.loc['StorageSize_MW']),       name=f"var.StorageDischarge.t{t}")
            self.var.H2Sales[t]                 = m.NumVar(lb=0, ub=int(zoneSetup.loc['StorageSize_MW']),       name=f"var.H2Sales.t{t}")                          
            self.var.ElBalance[t]               = m.NumVar(lb=0, ub=int(zoneSetup.loc['GridConnect_MW']),       name=f"var.ElBalance.t{t}")
            self.var.Curtailment[t]             = m.NumVar(lb=0, ub=m.infinity(),                               name=f"var.Curtailment.t{t}")
            self.var.BatteryStorage[t]          = m.NumVar(lb=0, ub=int(zoneSetup.loc['BatStorageSize_MWh']),   name=f"var.BatteryStorage.t{t}") 
            self.var.BatStorageCharge[t]        = m.NumVar(lb=-int(zoneSetup.loc['BatStorageSize_MW']), ub=0,   name=f"var.BatStorageCharge.t{t}")
            self.var.BatStorageDischarge[t]     = m.NumVar(lb=0, ub=int(zoneSetup.loc['BatStorageSize_MW']),    name=f"var.BatStorageDischarge.t{t}")
            self.var.Import[t]                  = m.NumVar(lb=0, ub=m.infinity(),                               name=f"var.Import.t{t}")
            self.var.ASRevenueBat[t]            = m.NumVar(lb=-m.infinity(), ub= m.infinity(),                  name=f"var.ASRevenueBat.t{t}")  
            self.var.ancServiceBat[t]           = {}

                  
            for AncS in self.sets.a:
                self.var.ancServiceBat[t][AncS] = m.NumVar(lb=0, ub=int(zoneSetup.loc['BatStorageSize_MW']) * self.data.ancService["Battery"][AncS], name=f"var.ancServiceBat.t{t}.{AncS}") 

            # -------- Hourly constraint       --------  
            # -------- Investment of both storage power and energy  --------
            self.cts.InvConstraintH2[t]         = m.Add(self.var.Storage[t]             <= self.var.InvestH2Storage,   name=f"cts.InvConstraintH2.t{t}")  
            self.cts.InvConstraintH2Charge1[t]  = m.Add(self.var.StorageCharge[t]       >= self.var.InvestH2Charge,    name=f"cts.InvConstraintH2Charge1.t{t}")
            self.cts.InvConstraintH2Charge2[t]  = m.Add(self.var.StorageDischarge[t]    <= self.var.InvestH2Discharge, name=f"cts.InvConstraintH2Charge2.t{t}") 
            self.cts.InvConstraintH2Charge3[t]  = m.Add(-self.var.InvestH2Charge        == self.var.InvestH2Discharge, name=f"cts.InvConstraintH2Charge3.t{t}")
            
            self.cts.InvConstraintBat[t]        = m.Add(self.var.BatteryStorage[t]      <= self.var.InvestBatStorage,   name=f"cts.InvConstraintBat.t{t}") 
            self.cts.InvConstraintBatCharge1[t] = m.Add(self.var.BatStorageCharge[t]    >= self.var.InvestBatCharge,    name=f"cts.InvConstraintBatCharge1.t{t}")
            self.cts.InvConstraintBatCharge2[t] = m.Add(self.var.BatStorageDischarge[t] <= self.var.InvestBatDischarge, name=f"cts.InvConstraintBatCharge2.t{t}") 
            self.cts.InvConstraintBatCharge3[t] = m.Add(-self.var.InvestBatCharge       == self.var.InvestBatDischarge, name=f"cts.InvConstraintBatCharge3.t{t}")
            
            
            # -------- Grid connection constraint, considering energy balances and ancillary service provision --------
            self.cts.InvConstraintGC1[t]        = m.Add(self.var.ElBalance[t] - self.var.Import[t] + sum(self.data.ancService['Up'][AncS] * self.U[uID].var.ancService[t][AncS] for uID, AncS in itertools.product(self.sets.u,self.sets.a)) + sum(self.var.ancServiceBat[t][AncS] * self.data.ancService['Up'][AncS] for AncS in self.sets.a) <= self.var.InvestGridConnect, 
                                                    name=f"cts.InvConstraintGC1.t{t}") 
            self.cts.InvConstraintGC2[t]        = m.Add(self.var.ElBalance[t] - self.var.Import[t] - sum(self.data.ancService['Dwn'][AncS] * self.U[uID].var.ancService[t][AncS] for uID, AncS in itertools.product(self.sets.u,self.sets.a)) - sum(self.var.ancServiceBat[t][AncS] * self.data.ancService['Dwn'][AncS] for AncS in self.sets.a) >= -self.var.InvestGridConnect, 
                                                    name=f"cts.InvConstraintGC2.t{t}") 

            # -------- H2 and electricity balances --------
            self.cts.H2Balance[t]               = m.Add(int(self.data.zoneData['Hydrogen_Demand'][t]) + self.var.H2Sales[t] == sum(self.U[uID].var.qNet[t] for uID in self.sets.u) + self.var.StorageCharge[t] + self.var.StorageDischarge[t],  
                                                    name=f"cts.H2Balance.t{t}") 
                                  
            self.cts.ElBalance[t]               = m.Add(self.var.ElBalance[t] == sum(self.U[uID].var.pNet[t] for uID in self.sets.u) - self.var.Curtailment[t] + self.var.Import[t] + self.var.BatStorageCharge[t] + self.var.BatStorageDischarge[t],  
                                                    name=f"cts.ElBalance.t{t}")                                     
            
            # -------- Ancillary service from Battery --------
            self.cts.ASRevenueBat[t]            = m.Add( self.var.ASRevenueBat[t] == sum(self.var.ancServiceBat[t][AncS] * self.data.zoneData[AncS+'_Price'][t] for AncS in self.sets.a), 
                                                    name=f"cts.ASRevenueBat.t{t}") 
    
            for AncS in self.sets.a:
                self.cts.ASProvidedBat[t]       = m.Add(self.var.ancServiceBat[t][AncS] <= self.data.ancService["Battery"][AncS] * self.var.InvestBatDischarge, 
                                                    name=f"cts.ASProvidedBat.AncS{AncS}.t{t}")                            
            self.cts.downRegBat[t]              = m.Add(self.var.InvestBatDischarge + (self.var.BatStorageCharge[t] + self.var.BatStorageDischarge[t]) >= sum(self.var.ancServiceBat[t][AncS] * self.data.ancService['Dwn'][AncS] for AncS in self.sets.a), 
                                                    name=f"cts.downRegBat.t{t}")    
            self.cts.upRegBat[t]                = m.Add(self.var.InvestBatDischarge - (self.var.BatStorageCharge[t] + self.var.BatStorageDischarge[t]) <= sum(self.var.ancServiceBat[t][AncS] * self.data.ancService['Up'][AncS] for AncS in self.sets.a), 
                                                    name=f"cts.upRegBat.t{t}")
            
            # -------- Storage energy balances --------
            if t == 0:
                self.cts.StorageBal[t]          = m.Add(self.var.Storage[t] == int(zoneSetup.loc['iniStorage_MWh']) - self.var.StorageCharge[t] - self.var.StorageDischarge[t],  
                                                    name=f"cts.StorageBal.t{t}")                                 
            else: self.cts.StorageBal[t]        = m.Add(self.var.Storage[t] == self.var.Storage[t-1] - self.var.StorageCharge[t] - self.var.StorageDischarge[t],  
                                                    name=f"cts.StorageBal.t{t}")    
             
            if t == 0:
                self.cts.BatSoCBal[t]           = m.Add(self.var.BatteryStorage[t] == int(zoneSetup.loc['BatiniStorage_MWh']) - (1-int(zoneSetup.loc['BatStoragePowerConsumption_MWh']))*self.var.BatStorageCharge[t] - self.var.BatStorageDischarge[t],  
                                                    name=f"cts.BatSoCBal.t{t}")               
            else: self.cts.BatSoCBal[t]         = m.Add(self.var.BatteryStorage[t] == self.var.BatteryStorage[t-1] - (1-int(zoneSetup.loc['BatStoragePowerConsumption_MWh']))*self.var.BatStorageCharge[t] - self.var.BatStorageDischarge[t],  
                                                    name=f"cts.BatSoCBal.t{t}") 

            for AncS in self.sets.a:
                self.cts.AncillaryServices1[t]  = m.Add(sum(self.U[uID].var.ancService[t][AncS] * self.data.ancService['Up'][AncS] for uID in self.sets.u) + self.var.ancServiceBat[t][AncS] * self.data.ancService['Up'][AncS] <= self.data.zoneData[AncS][t], 
                                                    name=f"cts.AncillaryServices1.AncS{AncS}.t{t}")
                self.cts.AncillaryServices2[t]  = m.Add(sum(self.U[uID].var.ancService[t][AncS] * self.data.ancService['Dwn'][AncS] for uID in self.sets.u) + self.var.ancServiceBat[t][AncS] * self.data.ancService['Dwn'][AncS] <= self.data.zoneData[AncS][t], 
                                                    name=f"cts.AncillaryServices2.AncS{AncS}.t{t}")
                

    def _build_objective(self):
        r"""
        Creates the overall objective function in which the optimization model 
        tries to minimize.
        """
        
        self.objective = self.model.Minimize((sum(self.U[u].var.totalCosts[t] - self.U[u].var.ASRevenue[t] for t, u in itertools.product(self.sets.t,self.sets.u))
                                              + sum(-self.var.H2Sales[t] * self.data.zoneData['H2_Marketprice'][t]
                                              + self.var.Curtailment[t] * self.data.zoneData['Elspot'][t]
                                              - int(zoneSetup.loc['StoragePowerConsumption_MWh']) * self.data.zoneData['Elspot'][t] * self.var.StorageCharge[t]
                                              + self.var.Import[t] * self.data.unitData[self.sets.u[0]]['pPOT']
                                              - (self.var.BatStorageCharge[t] + self.var.BatStorageDischarge[t]) * self.data.zoneData['Elspot'][t]
                                              - self.var.ASRevenueBat[t] for t in self.sets.t)
                                              + sum(self.U[u].var.InvCapacity * self.data.unitData[u]["InvestCost"] for u in self.sets.u)
                                              + self.var.InvestH2Storage * int(zoneSetup.loc['CapexH2Lager_DKK/MWh/y'])
                                              + self.var.InvestH2Discharge * int(zoneSetup.loc['CapexH2Lager_DKK/MW/y'])
                                              + self.var.InvestBatStorage * int(zoneSetup.loc['CapexBat_DKK/MWh/y'])
                                              + self.var.InvestBatDischarge * int(zoneSetup.loc['CapexBat_DKK/MW/y'])
                                              + self.var.InvestGridConnect * int(zoneSetup.loc['GridConnectCapex_DKK/MW/y']))/1e6)
    
    def export_results(self):
        r"""
        Exports the solution to a dataframe
        """
        
        dataDict = {}
        
        dataDict['Elspot']              = zoneData['Elspot']
        dataDict['Storage']             = {}
        dataDict['StorageCharge']       = {}
        dataDict['StorageDischarge']    = {}
        dataDict['BatteryStorage']      = {}
        dataDict['BatStorageCharge']    = {}
        dataDict['BatStorageDischarge'] = {}
        dataDict['BatteryStorage']      = {}
        dataDict['H2.Sold']             = {}
        dataDict['Import']              = {}
        dataDict['ElBalance']           = {}
        dataDict['Curtailment']         = {}
        dataDict['InvestH2Storage']     = {}
        dataDict['InvestBatStorage']    = {}
        dataDict['InvestH2Charge']      = {}
        dataDict['InvestBatCharge']     = {}
        dataDict['InvestGridConnect']   = {}
        dataDict['ASRevenueBat']        = {}
        
        for AncS in self.sets.a: 
            dataDict['ancServiceBat.'+AncS] = {}
            
        for u in self.sets.u:
            dataDict[u+'.P']             = {}
            dataDict[u+'.Q']             = {}
            dataDict[u+'.TotalCosts']    = {}
            dataDict[u+'.AS_Revenue']    = {}
      
            for AncS in self.sets.a: 
                dataDict[u+'.'+AncS]     = {}
                               
                                  
        for t in self.sets.t:
            dataDict['Storage'][t]                = self.var.Storage[t].solution_value()
            dataDict['StorageCharge'][t]          = self.var.StorageCharge[t].solution_value()
            dataDict['StorageDischarge'][t]       = self.var.StorageDischarge[t].solution_value()
            dataDict['BatteryStorage'][t]         = self.var.BatteryStorage[t].solution_value()
            dataDict['BatStorageCharge'][t]       = self.var.BatStorageCharge[t].solution_value()
            dataDict['BatStorageDischarge'][t]    = self.var.BatStorageDischarge[t].solution_value()
            dataDict['H2.Sold'][t]                = self.var.H2Sales[t].solution_value()
            dataDict['Import'][t]                 = self.var.Import[t].solution_value()
            dataDict['ElBalance'][t]              = self.var.ElBalance[t].solution_value()
            dataDict['Curtailment'][t]            = self.var.Curtailment[t].solution_value()
            dataDict['ASRevenueBat'][t]           = self.var.ASRevenueBat[t].solution_value()
            
            for AncS in self.sets.a:
                dataDict['ancServiceBat.'+AncS][t] = self.var.ancServiceBat[t][AncS].solution_value()
                          
            for u in self.sets.u:
                dataDict[u+'.P'][t]               = self.U[u].var.pNet[t].solution_value()
                dataDict[u+'.Q'][t]               = self.U[u].var.qNet[t].solution_value()
                dataDict[u+'.TotalCosts'][t]      = self.U[u].var.totalCosts[t].solution_value() 
                dataDict[u+'.AS_Revenue'][t]      = self.U[u].var.ASRevenue[t].solution_value()
                
                for AncS in self.sets.a: 
                    dataDict[u+'.'+AncS][t]       = self.U[u].var.ancService[t][AncS].solution_value()     
                        
        dataDict = pd.DataFrame.from_dict(dataDict)
                
        dataDict['H2_Revenue']                   = dataDict['H2.Sold'] * zoneData['H2_Marketprice']
        
        for u in self.sets.u:
            dataDict.loc[0,u+'.Investment size'] = self.U[u].var.InvCapacity.solution_value()
        
        dataDict.loc[0,'InvestH2Storage']        = BalOpt.var.InvestH2Storage.solution_value()
        dataDict.loc[0,'InvestH2Charge']         = BalOpt.var.InvestH2Charge.solution_value()
        dataDict.loc[0,'InvestBatStorage']       = BalOpt.var.InvestBatStorage.solution_value()
        dataDict.loc[0,'InvestBatCharge']        = BalOpt.var.InvestBatCharge.solution_value()        
        dataDict.loc[0,'InvestGridConnect']      = BalOpt.var.InvestGridConnect.solution_value()
        dataDict.loc[0,'TotalCost']              = BalOpt.model.Objective().Value()
        
        return dataDict  
            
#%%  Main initilization script

   
if __name__ == '__main__':
    path     = (os.path.dirname(os.path.abspath(__file__)))
    fullpath = (path+'\PtX_optimizer_Lin.xlsm')
    wb = xw.Book(fullpath)        
                
    grandTic = ticToc('Full model')
        
    # Import data from Excel 
    unitData        = wb.sheets['AvgUnit'].range('A1').options(pd.DataFrame, expand='table').value
    unitData        = unitData.drop('Unit',axis=1)
    zoneData        = wb.sheets['TimeSeries'].range('B1').options(pd.DataFrame, expand='table',index=False).value   
    zoneSetup       = wb.sheets['Storage'].range('A1').options(pd.DataFrame, expand='table').value
    ancS            = wb.sheets['Ancillary services'].range('A2').options(pd.DataFrame, expand='table').value
    ancS            = ancS[ancS['Active']>0]
    Main            = wb.sheets['Main'].range('A4').options(pd.DataFrame, expand='table').value
    tSet = list(range(len(zoneData)))
    uSet = list(unitData.columns) 
    aSet = list(ancS.index)

    # Initiate optimizer class
    BalOpt = BalanceOptimizer()
    
    # Load data
    BalOpt.load_data(zoneData, zoneSetup, unitData, ancS)
    BalOpt.load_sets(tSet,uSet,aSet)
    
    maxTime= int(Main.Input.MaxTime)
    
    # Build model
    tic = ticToc('Build model')
    BalOpt.build_model()
    tic.toc()
    
    # Solver time limit set in ms
    BalOpt.model.SetTimeLimit(1000*60*maxTime)    
    
    tic = ticToc('Optimize')
    BalOpt.optimize()
    tic.toc()   
          
    tic = ticToc('Output time')
    df = BalOpt.export_results()
    tic.toc()    
    grandTic.toc()

#%% Export the results to Excel via Xlwings
        
ws = wb.sheets["Results"]
ws.name = "Results"
ws.clear_contents()
ws.range("A1").value = df

#%% creates html figure via Plotly

import plotly.graph_objects as go
from plotly.subplots import make_subplots

ActiveUnits = unitData.loc['Capacity'][abs(unitData.loc['Capacity'])>0].index.tolist()
timeSet = list(range(len(df)))
x   = timeSet 

fig = make_subplots(rows=3, cols=2,specs=[[{"secondary_y": True},{"secondary_y": False}],[{"secondary_y": False},{"secondary_y": True}],[{"secondary_y": False},{"secondary_y": False}]],subplot_titles=["-Power production",'Operation (P/H2)','mFRR','aFRR','FCR'])

plotList = ActiveUnits

for u in ActiveUnits:
    fig.add_trace(go.Scatter(x=x, y=df[u+'.P'], hoverinfo='x+y',mode='lines',line_color=unitData[u].plotColor, name=u+'P',legendgroup='P'+u, stackgroup='one'),row=1,col=1) 
    fig.add_trace(go.Scatter(x=x, y=df[u+'.Q'], hoverinfo='x+y',mode='lines', name=u+'.H2',legendgroup='Q'+u, stackgroup='Two'),row=1,col=1) 
    fig.add_trace(go.Scatter(x=df[u+'.P'], y=df[u+'.Q'], hoverinfo='x+y', mode='markers',opacity=0.95,marker=dict(color='red'), name='PQ-plot',legendgroup='group'+u, showlegend=False),row=1,col=2) 
    fig.add_trace(go.Scatter(x=x, y=df[u+'.mFRR_Up'], hoverinfo='x+y',mode='lines',line_color=unitData[u].plotColor, name=u+'mFRR',legendgroup='P'+u, showlegend=False, stackgroup='one'),row=2,col=1) 
    fig.add_trace(go.Scatter(x=x, y=df[u+'.aFRR'], hoverinfo='x+y',mode='lines',line_color=unitData[u].plotColor, name=u+'aFRR',legendgroup='P'+u, showlegend=False, stackgroup='one'),row=2,col=2)
    fig.add_trace(go.Scatter(x=x, y=df[u+'.FCR'], hoverinfo='x+y',mode='lines',line_color=unitData[u].plotColor, name=u+'FCR',legendgroup='P'+u, showlegend=False, stackgroup='one'),row=3,col=1) 


fig.add_trace(go.Scatter(x=x, y=df['ancServiceBat.mFRR_Up'], hoverinfo='x+y',mode='lines',name='Battery.mFRR',legendgroup='Battery FCR', showlegend=True, stackgroup='one'),row=2,col=1) 
fig.add_trace(go.Scatter(x=x, y=df['ancServiceBat.aFRR'], hoverinfo='x+y',mode='lines', name='Battery.aFRR',legendgroup='Battery aFRR', showlegend=True, stackgroup='one'),row=2,col=2) 
fig.add_trace(go.Scatter(x=x, y=df['ancServiceBat.FCR'], hoverinfo='x+y',mode='lines', name='Battery.FCR',legendgroup='Battery mFRR', showlegend=True, stackgroup='one'),row=3,col=1)     
fig.add_trace(go.Scatter(x=x, y=df['Storage'], hoverinfo='x+y',mode='lines', name='Storage',legendgroup='Storage',line_color='blue'),row=1,col=1) 
fig.add_trace(go.Scatter(x=x, y=df['BatteryStorage'], hoverinfo='x+y',mode='lines', name='BatteryStorage',legendgroup='BatteryStorage',line_color='brown'),row=1,col=1)
fig.add_trace(go.Scatter(x=x, y=df['Elspot'], hoverinfo='x+y',mode='lines', name='Elspot',legendgroup='Elspot',line_color='black'),row=1,col=1,secondary_y=True) 

fig.update_xaxes(title_text='Time')
fig.update_yaxes(title_text='MW')
fig.update_yaxes(title_text="kr./MWh el", secondary_y=True)
fig['layout']['yaxis3']['title']='Hydrogen MW'
fig['layout']['xaxis2']['title']='Power MW'
fig.update_traces(marker_size=10)

fig.write_html('Energy.html', auto_open=True)
       